import numpy
from typing import Sequence, Optional
from dataclasses import dataclass, asdict
from tantri.dipoles.types import DipoleTO
from enum import Enum
import logging


# stuff for generating random dipoles from parameters

_logger = logging.getLogger(__name__)


class Orientation(str, Enum):
	# Enum for orientation, making string for json serialisation purposes
	#
	# Note that htis might not be infinitely extensible?
	# https://stackoverflow.com/questions/75040733/is-there-a-way-to-use-strenum-in-earlier-python-versions
	XY = "XY"
	Z = "Z"
	RANDOM = "RANDOM"


# A description of the parameters needed to generate random dipoles
@dataclass
class DipoleGenerationConfig:
	# note no actual checks anywhere that these are sensibly defined with min less than max etc.
	x_min: float
	x_max: float
	y_min: float
	y_max: float
	z_min: float
	z_max: float

	mag: float

	# these are log_10 of actual value
	w_log_min: float
	w_log_max: float

	orientation: Orientation

	dipole_count: int
	generation_seed: int

	def __post_init__(self):
		# This allows us to transparently set this with a string, while providing early warning of a type error
		self.orientation = Orientation(self.orientation)

	def as_dict(self) -> dict:
		return_dict = asdict(self)

		return_dict["orientation"] = return_dict["orientation"].value

		return return_dict


def make_dipoles(
	config: DipoleGenerationConfig,
	rng_override: Optional[numpy.random.Generator] = None,
) -> Sequence[DipoleTO]:

	if rng_override is None:
		_logger.info(
			f"Using the seed [{config.generation_seed}] provided by configuration for dipole generation"
		)
		rng = numpy.random.default_rng(config.generation_seed)
	else:
		_logger.info("Using overridden rng, of unknown seed")
		rng = rng_override

	dipoles = []

	for i in range(config.dipole_count):
		sx = rng.uniform(config.x_min, config.x_max)
		sy = rng.uniform(config.y_min, config.y_max)
		sz = rng.uniform(config.z_min, config.z_max)

		# orientation
		# 0, 1, 2
		# xy, z, random

		if config.orientation is Orientation.RANDOM:
			theta = numpy.arccos(2 * rng.random() - 1)
			phi = 2 * numpy.pi * rng.random()
		elif config.orientation is Orientation.Z:
			theta = 0
			phi = 0
		elif config.orientation is Orientation.XY:
			theta = numpy.pi / 2
			phi = 2 * numpy.pi * rng.random()
		else:
			raise ValueError(
				f"this shouldn't have happened, orientation index: {config}"
			)

		px = config.mag * numpy.cos(phi) * numpy.sin(theta)
		py = config.mag * numpy.sin(phi) * numpy.sin(theta)
		pz = config.mag * numpy.cos(theta)

		w = 10 ** rng.uniform(config.w_log_min, config.w_log_max)

		dipoles.append(
			DipoleTO(numpy.array([px, py, pz]), numpy.array([sx, sy, sz]), w)
		)

	return dipoles



import numpy
from dataclasses import dataclass, asdict


# Lazily just separating this from Dipole where there's additional cached stuff, this is just a thing
# we can use as a DTO For dipole info.
@dataclass
class DipoleTO:
	# assumed len 3
	p: numpy.ndarray
	s: numpy.ndarray

	# should be 1/tau up to some pis
	w: float

	def as_dict(self) -> dict:
		return asdict(self)



from typing import Sequence
from tantri.dipoles import DipoleTO


def row_to_dipole(input_dict: dict) -> DipoleTO:
	p = input_dict["p"]
	if len(p) != 3:
		raise ValueError(
			f"p parameter in input_dict [{input_dict}] does not have length 3"
		)
	s = input_dict["s"]
	if len(s) != 3:
		raise ValueError(
			f"s parameter in input_dict [{input_dict}] does not have length 3"
		)
	w = input_dict["w"]

	return DipoleTO(p, s, w)


def rows_to_dipoles(dot_dict_array: Sequence[dict]) -> Sequence[DipoleTO]:
	return [row_to_dipole(input_dict) for input_dict in dot_dict_array]



from dataclasses import dataclass
import numpy
import numpy.random
import typing
from enum import Enum
from tantri.dipoles.types import DipoleTO

import logging

_logger = logging.getLogger(__name__)


class DipoleMeasurementType(Enum):
	ELECTRIC_POTENTIAL = 1
	X_ELECTRIC_FIELD = 2


@dataclass(frozen=True)
class DotPosition:
	# assume len 3
	r: numpy.ndarray
	label: str


@dataclass
class WrappedDipole:
	# assumed len 3
	p: numpy.ndarray
	s: numpy.ndarray

	# should be 1/tau up to some pis
	w: float

	# For caching purposes tell each dipole where the dots are
	# TODO: This can be done better by only passing into the time series the non-repeated p s and w,
	# TODO: and then creating a new wrapper type to include all the cached stuff.
	# TODO: Realistically, the dot positions and measurement type data should live in the time series.
	dot_positions: typing.Sequence[DotPosition]

	measurement_type: DipoleMeasurementType

	def __post_init__(self) -> None:
		"""
		Coerce the inputs into numpy arrays.
		"""
		self.p = numpy.array(self.p)
		self.s = numpy.array(self.s)

		self.state = 1
		self.cache = {}
		for pos in self.dot_positions:
			if self.measurement_type is DipoleMeasurementType.ELECTRIC_POTENTIAL:
				self.cache[pos.label] = self.potential(pos)
			elif self.measurement_type is DipoleMeasurementType.X_ELECTRIC_FIELD:
				self.cache[pos.label] = self.e_field_x(pos)

	def potential(self, dot: DotPosition) -> float:
		# let's assume single dot at origin for now
		r_diff = self.s - dot.r
		return self.p.dot(r_diff) / (numpy.linalg.norm(r_diff) ** 3)

	def e_field_x(self, dot: DotPosition) -> float:
		# let's assume single dot at origin for now
		r_diff = self.s - dot.r
		norm = numpy.linalg.norm(r_diff)

		return (
			((3 * self.p.dot(r_diff) * r_diff / (norm**2)) - self.p) / (norm**3)
		)[0]

	def transition(
		self, dt: float, rng_to_use: typing.Optional[numpy.random.Generator] = None
	) -> typing.Dict[str, float]:
		rng: numpy.random.Generator
		if rng_to_use is None:
			rng = numpy.random.default_rng()
		else:
			rng = rng_to_use
		# if on average flipping often, then just return 0, basically this dipole has been all used up.
		# Facilitates going for different types of noise at very low freq?
		if dt * 10 >= 1 / self.w:
			# _logger.warning(
			# 	f"delta t {dt} is too long compared to dipole frequency {self.w}"
			# )
			self.state = rng.integers(0, 1, endpoint=True)
		else:
			prob = dt * self.w
			if rng.random() < prob:
				# _logger.debug("flip!")
				self.flip_state()
		return {k: self.state * v for k, v in self.cache.items()}

	def flip_state(self):
		self.state *= -1


def get_wrapped_dipoles(
	dipole_tos: typing.Sequence[DipoleTO],
	dots: typing.Sequence[DotPosition],
	measurement_type: DipoleMeasurementType,
) -> typing.Sequence[WrappedDipole]:
	return [
		WrappedDipole(
			p=dipole_to.p,
			s=dipole_to.s,
			w=dipole_to.w,
			dot_positions=dots,
			measurement_type=measurement_type,
		)
		for dipole_to in dipole_tos
	]


class DipoleTimeSeries:
	def __init__(
		self,
		dipoles: typing.Sequence[DipoleTO],
		dots: typing.Sequence[DotPosition],
		measurement_type: DipoleMeasurementType,
		dt: float,
		rng_to_use: typing.Optional[numpy.random.Generator] = None,
	):
		self.rng: numpy.random.Generator
		if rng_to_use is None:
			self.rng = numpy.random.default_rng()
		else:
			self.rng = rng_to_use

		self.dipoles = get_wrapped_dipoles(dipoles, dots, measurement_type)
		self.state = 0
		self.dt = dt

	def transition(self) -> typing.Dict[str, float]:
		new_vals = [dipole.transition(self.dt, self.rng) for dipole in self.dipoles]

		ret = {}
		for transition in new_vals:
			for k, v in transition.items():
				if k not in ret:
					ret[k] = v
				else:
					ret[k] += v
		return ret


__all__ = [
	"WrappedDipole",
	"DipoleTimeSeries",
	"DipoleTO",
]



import click
import logging
import tantri
import numpy
import tantri.cli.input_files.write_dipoles
import tantri.cli.file_importer
import tantri.dipoles
import json
import tantri.dipoles.generation
import pathlib
import tqdm  # type: ignore

_logger = logging.getLogger(__name__)

LOG_PATTERN = "%(asctime)s | %(levelname)-7s | %(name)s | %(message)s"


POTENTIAL = "electric-potential"
X_ELECTRIC_FIELD = "x-electric-field"


def _set_up_logging(filename):
	handlers = [logging.StreamHandler()]
	if filename is not None:
		handlers.append(logging.FileHandler(filename))
	logging.basicConfig(
		level=logging.DEBUG,
		format=LOG_PATTERN,
		handlers=handlers,
	)
	logging.getLogger("pdme").setLevel(logging.INFO)
	logging.captureWarnings(True)


@click.group()
@click.option(
	"--log", help="Enable logging to stream only", is_flag=True, default=False
)
@click.option("--log-file", help="A filename to use for logging (implies --log)")
@click.version_option(tantri.get_version())
def cli(log, log_file):
	"""Utilities for generating simulated TLS time series data."""
	if log or (log_file is not None):
		# log file has been provided, let's log
		_set_up_logging(log_file)


@cli.command()
@click.option(
	"--dipoles-file",
	default="dipoles.json",
	show_default=True,
	type=click.Path(exists=True, path_type=pathlib.Path),
	help="File with json array of dipoles",
)
@click.option(
	"--dots-file",
	default="dots.json",
	show_default=True,
	type=click.Path(exists=True, path_type=pathlib.Path),
	help="File with json array of dots",
)
@click.option(
	"--measurement-type",
	type=click.Choice([POTENTIAL, X_ELECTRIC_FIELD]),
	default=POTENTIAL,
	help="The type of measurement to simulate",
	show_default=True,
)
@click.option(
	"--delta-t",
	"-t",
	type=float,
	default=1,
	help="The delta t between time series iterations.",
	show_default=True,
)
@click.option(
	"--num-iterations",
	"-n",
	type=int,
	default=10,
	help="The number of iterations.",
	show_default=True,
)
@click.option(
	"--time-series-rng-seed",
	"-s",
	type=int,
	default=None,
	help="A seed to use to create an override default rng. You should set this.",
)
@click.option(
	"output_file",
	"-o",
	type=click.File("w"),
	help="The output file to write, in csv format",
	required=True,
)
@click.option("--header-row/--no-header-row", default=False, help="Write a header row")
def write_time_series(
	dipoles_file,
	dots_file,
	measurement_type,
	delta_t,
	num_iterations,
	time_series_rng_seed,
	output_file,
	header_row,
):
	"""
	Generate a time series for the passed in parameters.
	"""
	_logger.debug(
		f"Received parameters [dipoles_file: {dipoles_file}] and [dots_file: {dots_file}]"
	)
	dipoles = tantri.cli.file_importer.read_dipoles_json_file(dipoles_file)
	dots = tantri.cli.file_importer.read_dots_json_file(dots_file)

	if measurement_type == POTENTIAL:
		measurement_enum = tantri.dipoles.DipoleMeasurementType.ELECTRIC_POTENTIAL
		value_name = "V"
	elif measurement_type == X_ELECTRIC_FIELD:
		measurement_enum = tantri.dipoles.DipoleMeasurementType.X_ELECTRIC_FIELD
		value_name = "Ex"

	_logger.debug(f"Using measurement {measurement_enum.name}")
	labels = [dot.label for dot in dots]
	if header_row:
		value_labels = ", ".join([f"{value_name}_{label}" for label in labels])
		output_file.write(f"t (s), {value_labels}\n")

	_logger.debug(
		f"Going to simulate {num_iterations} iterations with a delta t of {delta_t}"
	)

	_logger.debug(f"Got seed {time_series_rng_seed}")
	if time_series_rng_seed is None:
		time_series = tantri.dipoles.DipoleTimeSeries(
			dipoles, dots, measurement_enum, delta_t
		)
	else:
		rng = numpy.random.default_rng(time_series_rng_seed)
		time_series = tantri.dipoles.DipoleTimeSeries(
			dipoles, dots, measurement_enum, delta_t, rng
		)

	for i in tqdm.trange(num_iterations):
		transition = time_series.transition()
		transition_values = ", ".join(str(transition[label]) for label in labels)
		output_file.write(f"{i * delta_t}, {transition_values}\n")


@cli.command()
@click.argument(
	"generation_config",
	type=click.Path(exists=True, path_type=pathlib.Path),
)
@click.argument(
	"output_file",
	type=click.File("w"),
)
@click.option(
	"--override-rng-seed", type=int, help="Seed to override the generation config spec."
)
def generate_dipoles(generation_config, output_file, override_rng_seed):
	"""Generate random dipoles as described by GENERATION_CONFIG and output to OUTPUT_FILE.

	GENERATION_CONFIG should be a JSON file that matches the appropriate spec, and OUTPUT_FILE will contain JSON formatted contents.
	OUTPUT_FILE will be overwritten, if it exists.

	If the --override-rng-seed is set, it's better to keep logs of the generation!
	"""
	_logger.debug(
		f"generate_dipoles was called, with config file {click.format_filename(generation_config)}"
	)
	_logger.debug(f"override_rng_seed: [{override_rng_seed}]")

	with open(generation_config, "r") as config_file:
		data = json.load(config_file)
	config = tantri.dipoles.generation.DipoleGenerationConfig(**data)

	override_rng = None
	if override_rng_seed is not None:
		_logger.info(f"Overriding the rng with a new one with seed {override_rng_seed}")
		override_rng = numpy.random.default_rng(override_rng_seed)
	_logger.debug(f"generating dipoles with config {config}...")
	generated = tantri.dipoles.generation.make_dipoles(config, override_rng)

	output_file.write(
		json.dumps(
			[g.as_dict() for g in generated],
			cls=tantri.cli.input_files.write_dipoles.NumpyEncoder,
		)
	)



import pathlib
import json
import logging
import tantri.cli.input_files
from typing import Sequence
import tantri.dipoles

_logger = logging.getLogger(__name__)

# note that json is insecure by default right?
# but don't worry for now
# TODO: if this ever matters, can improve file handling.


def read_data_from_filename(filename: pathlib.Path):
	try:
		with open(filename, "r") as file:
			return json.load(file)
	except Exception as e:
		_logger.error(
			f"failed to read the file {filename}, raising and aborting", exc_info=e
		)


def read_dots_json_file(filename: pathlib.Path) -> Sequence[tantri.dipoles.DotPosition]:
	data = read_data_from_filename(filename)
	return tantri.cli.input_files.rows_to_dots(data)


def read_dipoles_json_file(filename: pathlib.Path) -> Sequence[tantri.dipoles.DipoleTO]:
	data = read_data_from_filename(filename)
	return tantri.cli.input_files.rows_to_dipoles(data)



import numpy.random


class SimpleTelegraphTimeSeries:
	def __init__(
		self, rng_to_use: numpy.random.Generator, transition_probability: float
	):
		if transition_probability > 1 or transition_probability < 0:
			raise ValueError(
				f"Transition probabliity must be between 0 and 1, got {transition_probability} instead"
			)
		self.transition_probability = transition_probability

		self.rng: numpy.random.Generator
		if rng_to_use is None:
			self.rng = numpy.random.default_rng()
		else:
			self.rng = rng_to_use

		self.state = 0

	def transition(self) -> float:
		if self.rng.random() < self.transition_probability:
			self.state = 1 - self.state
		return self.state
